/*
	Implementation von files.h
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#ifdef sun 
#include<sys/types.h>
#endif

#include "files.h"


struct TFiles
/*
   Diese Struktur enthlt alle .lab - Dateien, sowie deren Anzahl
*/   
{
    int num, max;
    char ** files;
};  // TFiles


#ifdef sun     // Kompilieren, wenn auf einem tollen Solaris-System

int alphasort(const struct dirent **a, const struct dirent **b)
{
  return(strcmp((*a)->d_name, (*b)->d_name));
} //alphasort

int scandir(const char *dir, struct dirent ***namelist,
             int (*select)(const struct dirent *),
             int (*compar)(const struct dirent **, const struct dirent **))
/*
Diese Funktion braucht man nur auf tollen Solaris - Rechnern.
Auf allen anderen Betriebssystemen exisistiert sie in einer Bibliothek.
*/
{
   DIR *d;
   struct dirent *entry;
   register int i=0;
   size_t entrysize;

   if ((d=opendir(dir)) == NULL)
      return(-1);

   *namelist=NULL;
   while ((entry=readdir(d)) != NULL)
   {
     if (select == NULL || (select != NULL && (*select)(entry)))
     {
       *namelist=(struct dirent **)realloc((void *)(*namelist),
                  (size_t)((i+1)*sizeof(struct dirent *)));
         if (*namelist == NULL) return(-1);
         entrysize=sizeof(struct 
 dirent)-sizeof(entry->d_name)+strlen(entry->d_name)+1;
         (*namelist)[i]=(struct dirent *)malloc(entrysize);
         if ((*namelist)[i] == NULL) return(-1);
         memcpy((*namelist)[i], entry, entrysize);
         i++;
     }
   }
   if (closedir(d)) return(-1);
   if (i == 0) return(-1);
   if (compar != NULL)
     qsort((void *)(*namelist), (size_t)i, sizeof(struct dirent *),(void*) compar);

   return(i);
}  // scandir
#endif  // SunOS 


int FindLabFiles(pFiles * pF)
/*
   Findet alle Dateien im Verz., die ".lab" enthalten. Diese werden in einer Struktur gespeichert.
*/   
{
    struct dirent ** namelist;
    int n,i=0;
    n = scandir(".",&namelist,0,alphasort);
    if (n<0) return 1;
    (*pF) = (pFiles)calloc(1,sizeof(struct TFiles));  // Speicher fr den Struct reservieren
    (*pF)->files = (void*)calloc(n,sizeof(void*));
    (*pF)->num = -1;		// Damit beim ersten Aufruf von GetNext.. das 0. Element zurckgeg. wird
    while (n--)
    {
	if (strstr(namelist[n]->d_name,".lab"))
	{			// Scheint eine Labyrinth-Datei zu sein
	    (*pF)->files[i] = (char*)calloc(strlen(namelist[n]->d_name)+1,sizeof(char));
	    strcpy((*pF)->files[i],namelist[n]->d_name);	// Dann in die Liste aufnehmen
	    i++;
	}
	free(namelist[n]);  // Bereich fr diese Datei freigeben
    }
    (*pF)->files = realloc((*pF)->files,i*sizeof(void*));	// Wir brauchen nicht ganz so viel Platz...
    free(namelist);  // Ganze Liste freigeben
    (*pF)->max = i;
    return 0;	// Anzahl der Labyrinthe zurck
} // findLabFiles

char * GetNextFile(pFiles * pF)
/*
   Nchste Datei im Verzeichniss zurckgeben
*/   
{
    if ((*pF)->num<(*pF)->max-1)
    	return (*pF)->files[++(*pF)->num];
    return 0;
} // GetNextFile

char * GetPrevFile(pFiles * pF)
/*
   Vorherige Datei im Verzeichniss zurckgeben
*/   
{
    if ((*pF)->num>0)
	return (*pF)->files[--(*pF)->num];
    return 0;
} // GetPrevFile

void DeleteFileStruct(pFiles * pF)
/*
   Gibt den Speicher fr die Struktur, die die Dateien aus dem Verzeichniss beinhaltet frei
*/   
{
    int i;
    for (i=0; i <=(*pF)->max; i++) free((*pF)->files[i]);
    free((*pF)->files);
    free(pF);
} // DeleteFileStruct

int FileExists(char * filename)
/*
   Gibt 1 zurck, wenn Datei 'filename' vorhanden ist.
   sonst 0.
*/
{
	FILE * pFile;
	
	pFile = fopen(filename,"r");	// Testweise zum Lesen ffnen
	if (pFile==0) return 0;
	fclose(pFile);		// Datei gleich wieder schliessen
	
	return 1;			// 0, wenn Datei vorhanden
}  // FileExists
