/*
   Implementation von game.h
*/   

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "displaylab.h"
#include "labyrinth.h"
#include "files.h"


char * ChooseLabFile()
/*
   Lsst den Benutzer zwischen versch. Dateien auswhlen
*/   
{
    // Variablendefinitionen
    char c=0, * ch;
    char * filename;
    
    pFiles pF;		// Zeiger auf abstakten Datentyp
    FindLabFiles(&pF);	// Sucht alle Dateien .lab im Verzeichniss
    filename = GetNextFile(&pF);	// Erste Datei bekommen
    if (filename==0) return 0;
    PrintText(5,5,"Level auswhlen  (w: nchstes; s: vorheriges; l: Laden )");
    while ((c !='l')&&(c!='\n'))
    {
	PrintText(5,7,filename);
	c = GetKey();
	switch(c)
	{
	    case 'w':			// Nchste Datei auswhlen
		ch = GetNextFile(&pF);	
		if (ch!=0) filename = ch;
		break;
	    case 's':			// Vorherige Datei auswhlen
		ch = GetPrevFile(&pF);
		if (ch!=0) filename=ch;
    		break;		
	}  //switch
	PrintText(5,7,"                                             "); // Ausgabe des Dateinamens entf.
    } // while
    PrintText(5,5,"                                                                      ");
    return filename;		// Zeiger auf Dateinamen zurckgeben
}  // ChooseLabFile

void GetFileName(int x, int y, char * filename, int mustexist)
/*
   Benutzer an der angegebenen Stelle nach einem Dateinamen fragen
*/   
{
    char* fn;
    
   if (mustexist) 
   {  // Wenn die Datei existieren muss, kann auch der Auswahldialog verwendet werden.
       fn = ChooseLabFile();
       strcpy(filename,fn);
       return ;
   }
   PrintText(x,y,"Dateiname: ");
   ReadStr(filename);
   PrintText(x,y,"                                                ");
   PrintText(x,y+1,"                                              ");  // Ausgaben entfernen
} // GetFileName


void PrintInfo(int posX, int posY)
/*
	Gibt Infos fr den Benutzer aus (Tastenbelegung)
*/
{
    PrintText(posX,posY,  "a     Schritt nach links");
    PrintText(posX,posY+1,"s     Schritt nach unten");
    PrintText(posX,posY+2,"d     Schritt nach rechts");
    PrintText(posX,posY+3,"w     Schritt nach oben");
    PrintText(posX,posY+4,"u     Schritt rckgngig");
    PrintText(posX,posY+5,"q     Beenden");
    PrintText(posX,posY+6,"S     Spielstand speichern");
    PrintText(posX,posY+7,"L     Spielstand/Labyrinth laden");
    PrintText(posX,posY+8,"r     Labyrinth von vorn beginnen");
    PrintText(posX,posY+9,"g/G	 Gre des Labyrinths ndern");

    PrintText(posX,posY+11,"Schritt:");
} // PrintInfo

int GameLoop(pLab lab)
/*
   In dieser Funktion luft das eigentlich Spiel ab.
   Beendet der Spieler das Spiel mit 'q', wird 0 zurckgegeben;
   ansonsten die Anzhal der Schritte.
*/   
{
    // Variablendefinitionen
    char tmp[50];	// wird fr die Ausgabe der Schrittzahl verwendet
    char c, *ch; 		// fr Benutzereingabe
    static void (*drawfct)(int x, int y, char c);  // Zeichenfkt. Speichern
    if (drawfct == &PutIcon) SetDrawFct(&lab,&PutIcon);
    cls();					// Bildschirm leeren
    PrintInfo(2*sizeX(&lab)+5,1); 
    PrintLabyrinth(&lab);			// Labyrinth einmal ausgeben
   do 						// Schleife, in der das Spiel abluft
    {
	sprintf(tmp,"%4d",step(&lab));		// Dazu Zahl --> String (mit min. Breite 4)
	PrintText(2*sizeX(&lab)+15,12,tmp);	// Schrittzahl ausgeben
	c = GetKey(); 				// Zeichen von der Tastatur lesen
	switch(c)				// Auswertung der Eingabe
	{
	    case 'w': MovePlayer(&lab,'n'); break;	// oben
	    case 's': MovePlayer(&lab,'s'); break;	// unten
	    case 'a': MovePlayer(&lab,'w'); break;	// links
	    case 'd': MovePlayer(&lab,'e'); break;	// rechts
	    case 'u': UndoMove(&lab);	     break;	// undo
	    case 'r': return -1;	     break;	// restart		      
	    case 'L': return -2;	     break;	// laden		     
	    case 'U': ch = GetUndoBuffer(&lab);		// UndoBuffer ausgeben
		      ch[step(&lab)]=0;
		      PrintText(0,2*sizeY(&lab)+5,ch);
		break;	      
	    case 'S': GetFileName(0,2*sizeY(&lab)+4,tmp,0);		// speichern
		      if (strlen(tmp)>1) 
		      {
			if (!strstr(tmp,".lab")) strcat(tmp,".lab");
		      	SaveLevel(&lab,tmp); 
		      }
			break;
	    case 'G': SetDrawFct(&lab,&PutIconBig); cls(); PrintLabyrinth(&lab); PrintInfo(2*sizeX(&lab)+5,1);
		      drawfct = &PutIconBig;  // Zeichenfunktion PutIconBig() verwenden
		      break;
	    case 'g':		// Zeichenfunktion PutIcon() verwenden 
		      SetDrawFct(&lab,&PutIcon);    cls(); PrintLabyrinth(&lab); PrintInfo(2*sizeX(&lab)+5,1);
		      drawfct = &PutIcon;
		      break;
		      
	}
    } while ((c != 'q')&&(!GameWon(&lab)));	// Bis Benutzer 'q' eingibt oder gewonnen hat
    
    if (c=='q') return 0;		// Wurde durch den Spieler beendet, 0 zurckgeben
    return step(&lab);			// Spieler hat gewonnen? Schrittzahl zurckgeben! 
} // GameLoop





int MainLoop(char * filename)
/*
   Dieser Teil beinhaltet das Hauptmen, ist dafr verantwortlich, das
   Level-Dateien eingelesen werden und startet die eigentliche Spielschleife.
   Wird 0 zurckgegeben, hat der Benutzer ein Ende angefordert.
*/
{
    // Variablendefinitionen
    pLab  lab=0;		// Zeiger auf Struktur, die Labyrinthinfos enthlt
    int i=0;	
    char c;			// Eingabe des Benutzers	
    int quitgame=0;		// Will Benutzer beenden
    
    do				// Hauptschleife
    {
	if ((filename[0]==0)||(!FileExists(filename))||(i==-2))	// Wurde kein oder falscher Dateiname angegeben?
	{							// Oder Benutzer will laden  (i=-2)
	    cls();		
	    PrintText(0,0,"Labyrinth laden");
	    PrintText(0,1,"***************");
	    GetFileName(0,5,filename,1);   // Dann Benutzer fragen
	}
        LoadLevel(&lab, filename);		// Level aus Datei laden
       if (lab==0)
       {
	   fprintf(stderr,"Dateiformat fehlerhaft! Programm wird abgebrochen. Bitte eine Taste drcken.\n");
	   fflush(stdout);
	   c = GetKey();
	   return 0;
       }	   
	SetDrawFct(&lab,&PutIconBig);		// Ausgabefunktion zuweisen
    	i = GameLoop(lab);			// Hier luft das eigentliche Spiel ab!
	if (i==0) quitgame=1;
	if (i>0) 
	{
    	    PrintText(0,2*sizeY(&lab)+2,"Sie haben gewonnen!");	// Spieler hat gewonnen
       	    PrintText(0,2*sizeY(&lab)+3,"Mchten sie level (b)eenden, (w)iederholen, (l)evel wechseln oder (n)chstes spielen?");
    	    c = GetKey();			// Benutzereingabe
    	    switch(c)
    	    {
    		case 'b': i = 0; quitgame=1; 		      	break;  // Beenden
    		case 'l': i=-2;				 	break;	// Benutzer nach Dateinamen fragen
    		case 'w': i = -1; 			       	break;  // Wiederholen (eigentlich berfl.)
		case 'n': i = 0;	 			break; 
    		default: i = 0;				  	break;	// Bei falscher Eingabe beenden	  
    	    } // switch
	} // if
	DeleteLabyrinth(&lab);				// Labyrinth wieder freigeben
    } while (i!=0);	// bis Benutzer beendet (i = GameLoop(..), 0 bei Ende)
    
    if (quitgame) return 0;		// Beenden?
    return 1;				// evtl. nchstes Level 
} // MainLoop
