#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "labyrinth.h"
#include "files.h"

struct TLabyrinth
/*
   Struktur, welche die Daten des Labyrinths enthlt.
*/   
{
    int posX,posY;		// Position des Spielers
    int sizeX, sizeY;		// Gre des Labyrinths
    int step;			// Anzahl der Schritte
    int output;			// 1: Ausgabe soll erfolgen; 0 nicht
    char ** square;		// 2D-Array, in dem das Labyrinth gespeichert wird
    char * UndoBuff;		// Hier werden alle Zge gespeichert
    char * filename;		// Der Dateiname des Labyrinths
    void (*PutIconFct)(int x, int y, char c);  // Funktion zur Ausgabe von Zeichen
}; // struct TLabyrinth


int max(int i, int j)
/*
   Gibt die grere der beiden bergebenen Zahlen zurck (keine ffentliche Fkt)
*/   
{
    if (j > i) return j;
    else return i;
} // max


void DeleteLabyrinth(pLab * lab)
/*
   Speicher des Labyrinths wieder freigeben
*/   
{
    int i;
    free((*lab)->UndoBuff);
    for (i=0; i< (*lab)->sizeX; i++) free((*lab)->square[i]);
    free((*lab)->square);
    free(*lab);
    (*lab)=0;
} // DeleteLabyrinth

void CreateLabyrinth(pLab * lab, int sizeX, int sizeY)
{
    	// Variablendefinitionen
    	int i;

	*lab = (pLab)calloc(1,sizeof(struct TLabyrinth));  // Speicher fr Labyrinth-Struct reservieren
	(*lab)->output = 0;
	(*lab)->sizeX = sizeX;		// Gre in das Struct schreiben
	(*lab)->sizeY = sizeY;
	(*lab)->UndoBuff=(char*)calloc(1,sizeof(char));
	((*lab)->square) = (void*)calloc((*lab)->sizeX,sizeof(void*));		// Speicher fr 
	for (i=0; i < (*lab)->sizeX; i++)					// beide Dimensionen des
	    ((*lab)->square[i]) = (char*)calloc((*lab)->sizeY,sizeof(char));	// Feldes reservieren
}//CreateLabyrinth

void ResizeLabyrinth(pLab * lab, int sizeX, int sizeY)
{
    //Variablendefinitionen
    int i;

       ((*lab)->square)= (void*) realloc((*lab)->square, max((sizeX),(*lab)->sizeX) *sizeof(void*));
       for (i=0; i < max(sizeX,(*lab)->sizeX); i++)
       {
	   if (i>=(*lab)->sizeX) (*lab)->square[i] = (char*)calloc(1,sizeof(char));// Noch kein Feld vorhanden? erstellen!
	   ((*lab)->square[i]) = (char*)realloc((*lab)->square[i],max(sizeY,(*lab)->sizeY)*sizeof(char)); // Reservieren

	   if ((*lab)->sizeY<sizeY) memset((*lab)->square[i]+(*lab)->sizeY,' ',sizeY-(*lab)->sizeY); // Initialisieren
	   if ((*lab)->sizeX<i+1) memset((*lab)->square[i],' ',max((*lab)->sizeY,sizeY));  // Initialisieren
       }
	       
       (*lab)->sizeX = max(sizeX,(*lab)->sizeX);	// Labyrinthgre anpassen
       (*lab)->sizeY = max(sizeY,(*lab)->sizeY);
}//ResizeLabyrinth

void SetDrawFct(pLab * lab, void*fct)
/*
   Legt die Ausgabefunktion fest
   Wird keine festgelegt, wird das Labyrinth nicht ausgegeben (beim Bewegen).
*/   
{
    (*lab)->PutIconFct = fct;
    (*lab)->output = 1;
} // SetDrawFct

void PrintLabyrinth(pLab * lab)
/*
   Gibt das Labyrinth ber eine Ausgabefunktion aus ( "PutIconFct()" )
   Deren Adresse muss vorher in Tlabyrinth.PutIconFct bergeben werden.
*/   
{
    // Variablendefinitionen
    int x, y;		// Zhlvariablen fr X/Y - Koordinaten
    if (!(*lab)->output) return;  // Wurde schon eine Funktion zur Ausgabe festgelegt?
    for (y=0;y<(*lab)->sizeY; y++)		// Zeilen
	for (x=0;x<(*lab)->sizeX;x++)	// Spalten
	    (*(*lab)->PutIconFct)(x,y,(*lab)->square[x][y]);		// Zeichen von (X,Y) ausgeben

} // PrintLabyrinth

void PrintLabyrinthSimple(pLab * lab)
/*
   Gibt das Labyrinth einfach auf stdout aus.
*/   
{
    int x, y;
    for (y=0;y<(*lab)->sizeY;y++)
    {
	for (x=0;x<(*lab)->sizeX;x++)
	    printf("%c",(*lab)->square[x][y]);	// Zeilen / Spalten ausgeben
	printf("\n");				// Nach jeder Zeile Zeilenumbruch
    }
} // PrintLabyrinthSimple

// Funktionen, die Informationen ber das Labyrinth liefern...
int sizeX(pLab * lab)
{
    return (*lab)->sizeX;
}
int sizeY(pLab * lab)
{
    return (*lab)->sizeY;
}
int GetPosX(pLab * lab)
{
    return (*lab)->posX;
}
int GetPosY(pLab * lab)
{
    return (*lab)->posY;
}
int step(pLab * lab)
{
    return (*lab)->step;
}

void SetSquare(pLab * lab, char c, int posX, int posY)
{
    (*lab)->square[posX][posY] = c;
    return ;
}

char GetSquare(pLab * lab, int posX, int posY)
{
    return (*lab)->square[posX][posY];
}

char * GetUndoBuffer(pLab * lab)
{
    return (*lab)->UndoBuff;
}

int MovePlayer(pLab * lab, char way)
/*
   Diese Funktion bewegt die Spielfigur an die angegebene Position, sofern dies, abhngig von
   der aktuellen Position, zulssig ist.
   Kisten werden evtl. verschoben und eine Bildschirmausgabe wird vorgenommen, sofern eine Ausgabe-
   funktion festgelegt wurde.
   Die Funktion gibt 1 zurck, wenn eine Kiste verschoben wurde.
   Sonst 0
   Bei Fehler ist der Rckgabewert 5
*/   
{
    // Variablendefinitionen
    int boxmoved=0;	// Wird 1, wenn Kiste verschoben wurde
    int x, y;
    int KposX, KposY;	// Zielpositionen fr die Kiste

     
    x = (*lab)->posX;
    y = (*lab)->posY;
    
    switch(way)
    {
	case 'n':
	    y-- ; break;
	case 'e':
	    x++; break;
	case 's':
	    y++; break;
	case 'w':
	    x--; break;
    }
   
    switch((*lab)->square[x][y])		// Zielfeld untersuchen
    {
    	case ' ': 		// Schritt auf leeres Feld oder auf Endfeld
	case 'E':    
    
	    if ((*lab)->square[x][y]==' ') (*lab)->square[x][y]='P';	// Schritt auf leeres Feld 
	    if ((*lab)->square[x][y]=='E') (*lab)->square[x][y]='G';	// oder Endfeld ?
	    switch ((*lab)->square[(*lab)->posX][(*lab)->posY])		// vorheriges Feld ?
	    {
    		case 'P':			// Spieler stand auf leerem Feld
    		    (*lab)->square[(*lab)->posX][(*lab)->posY] = ' ';
    		    break;
    		case 'G':			// Spieler stand auf einem Zielfeld
    		    (*lab)->square[(*lab)->posX][(*lab)->posY] = 'E';
    		    break;
    		default:
    		    fprintf(stderr,"Fehler: (labyrinth.h) Fkt. MovePlayer: Zeichen fr Spieler (%c) nicht erkannt!\n",(*lab)->square[(*lab)->posX][(*lab)->posY]);
    		    return 5;
    	    }
	    if ((*lab)->output)
	    {
	    	(*(*lab)->PutIconFct)(x,y,(*lab)->square[x][y]);		// Genderte Zeichen ausgeben
	    	(*(*lab)->PutIconFct)((*lab)->posX,(*lab)->posY,(*lab)->square[(*lab)->posX][(*lab)->posY]);
	    }
	    
    	    (*lab)->posX = x;		// Positionen aktualisieren
    	    (*lab)->posY = y;
    	    break;
	    
    	case 'K':		// Spieler zieht auf eine Kiste / Zielfeld mit Kiste
	case 'Z':   
		KposX = x + x - ((*lab)->posX);	// Kistenpositionen ausrechnen
		KposY = y + y - ((*lab)->posY);
		if (((*lab)->square[KposX][KposY]==' ')||((*lab)->square[KposX][KposY]=='E'))
		{
		    if ((*lab)->square[KposX][KposY]==' ') 
		    {
			(*lab)->square[KposX][KposY]='K';	// Kiste auf leeres Feld
			boxmoved = 1;		// Rckgabewert der Fkt
		    }
		    if ((*lab)->square[KposX][KposY]=='E')
		    {
			(*lab)->square[KposX][KposY]='Z';	// Kiste auf Zielfeld
			boxmoved=1;		// Rckgabewert der Fkt
		    }
		    if ((*lab)->square[x][y]=='K') (*lab)->square[x][y]='P';			// Kiste stand auf norm. Feld
		    if ((*lab)->square[x][y]=='Z') (*lab)->square[x][y]='G';			// Kiste stand auf Zielfeld
		    if ((*lab)->square[(*lab)->posX][(*lab)->posY]=='P') (*lab)->square[(*lab)->posX][(*lab)->posY]=' '; // Spieler stand auf
		    if ((*lab)->square[(*lab)->posX][(*lab)->posY]=='G') (*lab)->square[(*lab)->posX][(*lab)->posY]='E'; //   Ende ?
		    if ((*lab)->output)
		    {
		    	(*(*lab)->PutIconFct)(x,y,(*lab)->square[x][y]);	//genderte Zeichen ausgeben
		    	(*(*lab)->PutIconFct)((*lab)->posX,(*lab)->posY,(*lab)->square[(*lab)->posX][(*lab)->posY]);
		    	(*(*lab)->PutIconFct)(KposX,KposY,(*lab)->square[KposX][KposY]);
		    }
		    (*lab)->posX = x;		// Positionen aktualisieren
		    (*lab)->posY = y;
		
	    }		// Hoffentlich alle Bewegungen richtig ausgefhrt...
	    
	    break;
    } // switch	
    
    if (((*lab)->posX==x)&&((*lab)->posY==y))		// Bewegung ausgefhrt?
    {
	if (boxmoved) way=toupper(way);		// Kiste verschoben? Dann Grossbuchstaben im Undo-Buffer
	(*lab)->UndoBuff = (char*)realloc((*lab)->UndoBuff,((*lab)->step+1)*sizeof(char)); // Undo-Buffer vergrern
	(*lab)->UndoBuff[(*lab)->step]=way; // Richtung in den Undo-Buffer schreiben
	(*lab)->step++;				// Schrittzahl um 1 erhhen
    }
    
    return boxmoved;
} // MovePlayer

void UndoMove(pLab * lab)
/*
   Macht einen Zug rckgngig
*/   
{
    int dx=0, dy=0;	// Vernderung der Position im Labyrinth (Spieler)
    int BoxMoved=0;
    
    if ((*lab)->step<1) return;  // Wurde berhaupt schon ein Schritt gemacht? 
    ((*lab)->step)--;			// Schrittzahl um 1 erniedrigen
    if (isupper((int)(*lab)->UndoBuff[(*lab)->step])) BoxMoved = 1;// Ist UndoBuff Grobuchst. ==> Kiste wurde verrsch.
    (*lab)->UndoBuff[(*lab)->step] = tolower((*lab)->UndoBuff[(*lab)->step]);	// Kleinbuchstaben machen
    switch((*lab)->UndoBuff[(*lab)->step])		// Richtung feststellen
    {
	case 'n':	dy = 1; break;		//Noden
	case 's':	dy = -1; break;		//Sden
	case 'e':	dx = -1; break;		//Osten
	case 'w':	dx = 1; break;		// Westen
    }
    if (BoxMoved)		// Es wurde eine Kiste bewegt?
    {
    	if (((*lab)->square[((*lab)->posX)-dx][((*lab)->posY)-dy]=='K')  // Tastchlich Kiste nebenan?
		||((*lab)->square[((*lab)->posX)-dx][((*lab)->posY)-dy]=='Z'))   // also entw. K od. Z
	    {	
		if ((*lab)->square[((*lab)->posX)-dx][((*lab)->posY)-dy]=='K')	// Kiste stand auf leerem Feld
   	 	    (*lab)->square[((*lab)->posX)-dx][((*lab)->posY)-dy]=' ';	// also wird es wieder leer!
   	  	if ((*lab)->square[((*lab)->posX)-dx][((*lab)->posY)-dy]=='Z')	// Kiste stand auf Zielfeld
		    (*lab)->square[((*lab)->posX)-dx][((*lab)->posY)-dy]='E';	// wird also Endfeld
		if ((*lab)->square[(*lab)->posX][(*lab)->posY]=='P') // Spieler stand auf leerem Feld
		    (*lab)->square[(*lab)->posX][(*lab)->posY]='K';  // Also Kiste jetzt auf leerem Feld
		if ((*lab)->square[(*lab)->posX][(*lab)->posY]=='G') // Spieler stand auf Zielfeld?
		    (*lab)->square[(*lab)->posX][(*lab)->posY]='Z'; // Also Kiste jetzt auf Zielfeld
    	}
	else fprintf(stderr,"Fehler (labyrinth.h) Fkt. UndoMove: Kiste erwartet (statt '%c')\n",(*lab)->square[(*lab)->posX-dx][(*lab)->posY-dy]);
    }
    else	// Keine Kiste verschoben?
    {
	if ((*lab)->square[(*lab)->posX][(*lab)->posY]=='P')   // Spieler stand auf leerem Feld
	    (*lab)->square[(*lab)->posX][(*lab)->posY] = ' ';	 // also wieder leeres Feld
	if ((*lab)->square[(*lab)->posX][(*lab)->posY]=='G')   // Spieler stand auf Zielfeld
	    (*lab)->square[(*lab)->posX][(*lab)->posY] = 'E';	// also wieder Zielfeld
    }
    // Fr beide Flle (Kiste verschoben oder auch nicht): Spieler bewegen
    if ((*lab)->square[((*lab)->posX)+dx][((*lab)->posY)+dy]==' ')  // Neues Spielerfeld ist leer 
	(*lab)->square[((*lab)->posX)+dx][((*lab)->posY)+dy]='P';   // Also Spieler auf leerem Feld
    if ((*lab)->square[((*lab)->posX)+dx][((*lab)->posY)+dy]=='E') // Neues Spielerfeld ist Zielfeld
	(*lab)->square[((*lab)->posX)+dx][((*lab)->posY)+dy]='G';// Also Spieler auf Zielfeld
    
    // Ausgabe der neuen Felder (Kiste vorher, Spieler vorher, Spieler nachher)
    (*(*lab)->PutIconFct)((*lab)->posX-dx,((*lab)->posY)-dy,(*lab)->square[((*lab)->posX)-dx][((*lab)->posY)-dy]);	
    (*(*lab)->PutIconFct)((*lab)->posX,((*lab)->posY),(*lab)->square[(*lab)->posX][(*lab)->posY]);
    (*(*lab)->PutIconFct)(((*lab)->posX)+dx,((*lab)->posY)+dy,(*lab)->square[((*lab)->posX)+dx][((*lab)->posY)+dy]);
    
    (*lab)->posX += dx;		// Position aktualisieren
    (*lab)->posY += dy;
    
    (*lab)->UndoBuff = (char*)realloc((*lab)->UndoBuff,((*lab)->step+1)*sizeof(char)); // Speicher freigeben
    
    return ;
}  // UndoMove

int GameWon(pLab * lab)
/*
   Gibt 1 zurck, wenn sich keine Kiste mehr im Feld befindet.
*/   
{
    int i;		// Zhlvariable
    char * c;
    for (i=0;i<((*lab)->sizeX);i++)
    {
	c = strchr((*lab)->square[i],'K');		// Vgl. Zeilenweise mit 'K'
	if (c!=0) return 0;			// wird ein 'K' gefunden => Game not won!
    }
    return 1;
} // GameWon




void LoadLevel(pLab * lab, char * filename)
/*
   Funktion ldt Level aus einer Datei, die im entsprechnenden
   Format vorliegt.
*/   
{
    	// Variablendefinitionen
	FILE * pFile;		// Dateizeiger
	int i, j;		// Zhlvariablen fr Zeilen/Spalten
	char * buffer;		// Lesepuffer (muss vor dem Schreiben ins Array umgedreht werden)
	


	if (!FileExists(filename)) 
	{			// Existiert Datei berhaupt?
	    free(*lab);		// Nein? Dann Speicher wieder freigeben und abbrechen!
	    return ;		
	}
	pFile = fopen(filename,"r");			// Datei zum Lesen ffnen
	buffer = (char*)calloc(10,sizeof(char));	// Puffer, um die Gren zu lesen
	fgets(buffer,10*sizeof(char),pFile);		// fscanf geht nicht, weil es die Leerzeichen wegliest
	if (sscanf(buffer,"%d %d\n",&i,&j)!=2) 	
	{		// Werden in Datei zwei Dezimalen fr Gre angegeben?
	    free(*lab);	// Nein? Dann Speicher freigeben und abbrechen!
	    return ;
	}
	free(buffer);					// Puffer wieder freigeben

	// Speicherreservierungen
	CreateLabyrinth(lab,i,j);   // Speicher fr das Labyrinth bekommen
	(*lab)->filename=filename;
	buffer = (char*)calloc((*lab)->sizeX+3,sizeof(char));	// Speicher fr den tmp. Lesepuffer reservieren

	// Aus Datei lesen
	for (i=0; i < ((*lab)->sizeY); i++)
	{
		if (fgets(buffer,sizeof(char)*(((*lab)->sizeX)+3),pFile))		// in buffer lesen
		{
			if (buffer[(*lab)->sizeX]=='\n') 		// Zeilenumbruch entfernen
		    	buffer[(*lab)->sizeX]=0;
			else 
			{
			    fprintf(stderr,"Fehler: (labyrinth.h) Fkt. LoadLevel: Fehlerhaftes Dateiformat (keine Zeilenumbrche)!\n");
			    DeleteLabyrinth(lab);
			    return ; // Keine Zeilenumbrche vorhanden? Fehler!
			}
			j = strcspn(buffer,"PG");	// Ist der Spieler in dieser Zeile?
			if (j<((*lab)->sizeX))
			{
			    if (((*lab)->posX==0)&&((*lab)->posY==0))  // nur EIN Spieler
			    {
				(*lab)->posX = j;  // Positionen speichern 
				(*lab)->posY = i;
			    }
			    else		// Spieler schon vorhanden
			    {
				fprintf(stderr,"Fehler: (labyrinth.h) Fkt. LoadLevel: Nur ein Spieler erlaubt!\n");
				DeleteLabyrinth(lab);
				return;
			    }
			}
			
			for (j=0; j<((*lab)->sizeX); j++) 
			{
		    		(*lab)->square[j][i] = buffer[j]; // X-Y vertauschen (damit square[x][y] vorliegt)
			}

		} // if fgets(..)
		else 	// Unerwartetes Dateiende
		{
		    fprintf(stderr,"Fehler: (labyrinth.h) Fkt. LoadLevel: Unerwartetes Dateiende!\n");
		    DeleteLabyrinth(lab);
		    return ;
		}
	} // for

	free(buffer);					// buffer wieder freigeben
	fclose(pFile);					// Datei schliessen
	
	if (((*lab)->posX==0)||((*lab)->posY==0))	// Auch ein Spieler im Labyrinth
	{
	    fprintf(stderr,"Fehler: (labyrinth.h) Fkt. LoadLevel: Kein Spieler gefunden!\n");
	    DeleteLabyrinth(lab);
	    return ;
	}
	return;
}  // LoadLevel


void SaveLevel(pLab * lab, char * filename)
/*
   Speichert ein Level in einer Datei
   (entweder zum Speichern eines Spielstandes oder zum Generieren von Levels)
*/   
{
    	// Variablendefinitionen
	FILE * pFile;		// Dateizeiger
	int x, y;		// Zhlvariablen fr Zeilen/Spalten
	
	pFile = fopen(filename,"w");			// zum Schreiben ffnen
	fprintf(pFile,"%d %d\n",(*lab)->sizeX,(*lab)->sizeY);
	for (y=0; y < ((*lab)->sizeY); y++)	// Zeilen
	{
		for (x=0; x < ((*lab)->sizeX); x++) // Spalten			
			fprintf(pFile,"%c",(*lab)->square[x][y]);		// Labyrinth zeichenweise schreiben
		fprintf(pFile,"\n");				// Nach jeder Zeile Zeilenumbruch
	}
	fclose(pFile);
} // SaveLevel

